#!/usr/bin/env python3

import json
import sqlite3
from collections import defaultdict

logfile_path = "result.jsonl"
database_path = "result.db"

# Converts a file in JSON Lines text file format into an SQLite database.
# The resulting table will have a column for each property that appears
# in the logfile. The file tries to preserve the data types of the
# properties. Missing properties will become NULL values in the table.

# 1st pass: collect properties and create table schema
properties = defaultdict(set)  # Set stores the types of each property

# Open file and determine types of properties
with open(logfile_path, "r") as file:
    for line in file:
        entry = json.loads(line.strip())
        for key, value in entry.items():
            if isinstance(value, bool):
                properties[key].add("BOOLEAN")
            elif isinstance(value, int):
                properties[key].add("INTEGER")
            elif isinstance(value, float):
                properties[key].add("REAL")
            else:
                properties[key].add("TEXT")

# define SQL types
columns = {}
for key, types in properties.items():
    if "TEXT" in types:
        columns[key] = "TEXT"
    elif "REAL" in types:
        columns[key] = "REAL"
    elif "INTEGER" in types:
        columns[key] = "INTEGER"
    elif "BOOLEAN" in types:
        columns[key] = "BOOLEAN"
    else:
        columns[key] = "TEXT"

create_table_query = f"CREATE TABLE IF NOT EXISTS log ({', '.join([f'{col} {dtype}' for col, dtype in columns.items()])});"

# Create database connection and table
conn = sqlite3.connect(database_path)
cur = conn.cursor()
cur.execute(create_table_query)
conn.commit()

# 2nd pass: insert data into database
with open(logfile_path, "r") as file:
    for line in file:
        entry = json.loads(line.strip())

        # prepare SQL statement to insert the data
        columns_str = ", ".join(columns.keys())
        placeholders = ", ".join([f":{col}" for col in columns.keys()])
        insert_query = f"INSERT INTO log ({columns_str}) VALUES ({placeholders})"

        # Prepare data with NULL for missing values
        data = {col: entry.get(col) for col in columns.keys()}

        # insert data into database
        cur.execute(insert_query, data)

# save changes and close the file
conn.commit()
conn.close()

print("Data successfully written to the SQLite database.")
